/* interp.c    interpreter i-code	9 Jan 97

	som-v16 add file i/o 			7 Nov 2005
	som-v16a  optimised speed 		25 Dec 2006 (Christmas!)
 */

#include "compile.h"

// stack macro
#define   TOS       SS[sp]
#define   NEXT      SS[sp-1]
#define   pop2		sp -= 2
#define	  pop1      sp--

//PRIVATE void push(int x) { sp++; SS[sp] = x; }
//PRIVATE int pop(void){ int x; x = TOS; pop1; return x; }
// inline is 4x faster than push/pop functions
#define   push(x) 	sp++; SS[sp] = (x)
#define   pop()		SS[sp]; sp--
// assume local var a
#define   alu(result) 	a = (result); pop1; TOS = a


/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..80   tokstring-som
*/

extern int load_ads;		// base ads of fun "loadfile"

int M[MAXMEM];				// data and code segment
int SS[MAXSS];				// stack segment
int *CS = &M[MAXDS];		// code segment
int DS = MAXSYS;			// current free data segment

PRIVATE FILE *filep[MAXFILE];	// store file pointer

PRIVATE int fp = 1, sp = 0;	// current state of computation
PRIVATE int opcode,data;	// current instruction
PRIVATE int IP;				// for error report

void initfile(void){
	int i;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
}
// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}

/*
// !!! debug

int callstk[200], callstkptr = 0;
void pushcallstk(int x){
	callstkptr++;
	callstk[callstkptr] = x;
}
void popcallstk(void){ callstkptr--; }

void dumpcallstk(void){
	int i;
	printf("call stack :");
	for(i=1; i<=callstkptr; i++)
		printf("%d ",callstk[i]);
	nl();
}

PRIVATE int traceflag = 0;

extern FILE *FO;
extern int lv;

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
	prCode(ip);
	printf(" |");
	a = sp - 10;
	if( a < 0 ) a = 0;

	for(i=a; i<=sp; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}

void dumpDS(void){
	int i;
	for(i=1; i<10; i++)	printf("%d:%d ",i,M[i]);
	printf("\n");
}
*/

PRIVATE void error(char *mess){
	fprintf(stderr,"runtime error: %s at ip %d\n",mess,IP);
	exit(-1);
}

// fgets(fp) and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		error("alloc: out of memory");
	DS += a;
	return b;
}
// exec syscall n
PRIVATE void xsys(int n){
	int a, f;
	char name[NAMELEN];

	if(n != 3){ a = pop(); }		// TOS arg.
	switch(n){
	case 1: printf("%d",a); break; 	// pr int
	case 2: printf("%c",a); break;	// pr char
	case 3: push(getchar()); break;	// getchar
	case 4: xgets(stdin,a); break;	// gets
	case 5:							// fopen fname mode
		f = pop();					// fname
		strunpack(name,f);
		f = newfp();
		if(f == 0)
			error("cannot open, too many files");
		switch(a){					// mode
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"rw"); break;
		default: error("unknown file mode");
		}
		push(f);					// return file pointer
		break;
	case 6:							// fclose fp
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:							// fprint fp num
		f = pop();
		fprintf(filep[f],"%d",a);
		break;
	case 8:							// fprintc fp c
		f = pop();
		fprintf(filep[f],"%c",a);
		break;
	case 9:	push(fgetc(filep[a]));  // fgetc fp
		break;
	case 10:						// fgets fp inbuf
		f = pop();
		xgets(filep[f],a);
		break;
//	case 11:						// loadfile
//		strunpack(name,a);			// file name in som-string
//		loadfile(name);
//		break;
	case 12:						// eval from som
		CS = &M[M[CS_ADS]];			// switch to user space
		eval(a);
		CS = &M[MAXDS];				// switch back
		break;
//	case 13: runflag = 0;  			// halt, this is in eval()
	case 14: push(xalloc(a)); break;
	case 15:						// exec load from user space
		push(a);
		CS = &M[MAXDS];				// to system
		eval(load_ads);
		CS = &M[M[CS_ADS]];			// switch to user
		break;
	default: error("undefined syscall");
	}
}

void eval(int ref){
//	int limit = 200000;
	int ip, runflag;
	int a, b, low, hi;

	ip = ref;
	runflag = 1;
    while(runflag ) {
//		if( (limit--) < 0 ) break;
//		if (traceflag) trace();
		IP = ip;
		opcode = CS[ip] & 255;
		data = CS[ip] >> 8 ;
		if (opcode > EOC)
			error("undefined code");

// ---------- exec ---------

	switch(opcode){
	case icAdd: alu(NEXT + TOS); ip++; break;
	case icSub: alu(NEXT - TOS); ip++; break;
	case icMul: alu(NEXT * TOS); ip++; break;
	case icDiv: alu(NEXT / TOS); ip++; break;
	case icBand: alu(NEXT & TOS); ip++; break;
	case icBor: alu(NEXT | TOS); ip++; break;
	case icBxor: alu(NEXT ^ TOS); ip++; break;
	case icMod: alu(NEXT % TOS); ip++; break;
	case icShl: alu(NEXT << TOS); ip++; break;
	case icShr: alu(NEXT >> TOS); ip++; break;
	case icEq: alu(NEXT == TOS); ip++; break;
	case icNe: alu(NEXT != TOS); ip++; break;
	case icLt: alu(NEXT < TOS); ip++; break;
	case icLe: alu(NEXT <= TOS); ip++; break;
	case icGt: alu(NEXT > TOS); ip++; break;
	case icGe: alu(NEXT >= TOS); ip++; break;

	case icNot: TOS = (TOS == FALSE); ip++; break;
	case icInc: SS[fp-data]++; ip++; break;
	case icDec: SS[fp-data]--; ip++; break;
	case icLit:
	case icAds:	push(data);	ip++; break;
	case icLd:	push( M[data]);	ip++; break;
	case icSt:
		M[data] = TOS;
		pop1;
		ip++;
		break;
	case icPut:
		SS[fp-data] = pop();
		ip++;
		break;
	case icGet:
		push(SS[fp-data]);
		ip++;
		break;
	case icLdx:
		a = NEXT + TOS;
		pop1;
		TOS = M[a];
		ip++;
		break;
	case icStx:
		a = pop();
		b = NEXT + TOS;
		pop2;
		M[b] = a;
		ip++;
		break;
	case icArray:
		a = pop();
		push(xalloc(a));
		ip++;
		break;
	case icJmp:
		ip += data;
		break;
	case icJf:
		a = pop();
		if( a == FALSE) ip += data;	else ip++;
		break;
	case icJt:
		a = pop();
		if( a != FALSE) ip += data;	else ip++;
		break;
	case icCall:
		a = CS[data] >> 8;		// get lv
		if (sp + a + FUDGE > MAXSS)
			error("stack overflow");
		SS[sp+a] = fp;
		fp = sp + a;
		sp = fp + 1;
		SS[sp] = ip + 1;
		ip = data + 1;
//		pushcallstk(data);	// !!! debug
		break;
	case icRet:
	// return TOS to caller if current stack is not empty
		ip = SS[fp+1];
		if ( sp > fp+1 ){
			a = TOS;
			sp = fp - data + 1;
			fp = SS[fp];
			SS[sp] = a;
		}else{
			sp = fp - data;
			fp = SS[fp];
		}
//		popcallstk();		// !!! debug
		break;
	case icCase:
		low = CS[ip+1] >> 8;
		hi = CS[ip+2] >> 8;
		a = pop();
		if( a < low || a > hi )
			ip = ip+3;
		else
			ip += a-low+4;
		break;
	case icSys:
		if( data == 13 ) runflag = 0;
		else xsys(data);
		ip++;
		break;
	case icEnd:	runflag = 0; break;
	default: error("undefined opcode");
    }
// ----- end exec ----
	}
}


